@@ -1,6 +1,6 @@
 # Copyright (C) 2014,2015 VA Linux Systems Japan K.K.
 # Copyright (C) 2014,2015 YAMAMOTO Takashi <yamamoto at valinux co jp>
-# Copyright (C) 2014 Fumihiko Kakuma <kakuma at valinux co jp>
+# Copyright (C) 2014,2015 Fumihiko Kakuma <kakuma at valinux co jp>
 # All Rights Reserved.
 #
 # Based on openvswitch agent.
@@ -23,6 +23,8 @@
 import time
 
 from oslo_config import cfg
+from oslo_log import helpers as log_helpers
+from oslo_log import log as logging
 import oslo_messaging
 from ryu.app.ofctl import api as ryu_api
 from ryu.base import app_manager
@@ -38,11 +40,9 @@ from neutron.agent.linux import ovs_lib
 from neutron.agent import rpc as agent_rpc
 from neutron.agent import securitygroups_rpc as sg_rpc
 from neutron.common import constants as n_const
-from neutron.common import log
 from neutron.common import topics
 from neutron.common import utils as n_utils
 from neutron import context
-from neutron.openstack.common import log as logging
 from neutron.openstack.common import loopingcall
 from neutron.plugins.common import constants as p_const
 from neutron.plugins.openvswitch.common import constants
@@ -51,6 +51,7 @@ from networking_ofagent.i18n import _LE, _LI, _LW
 from networking_ofagent.plugins.ofagent.agent import arp_lib
 from networking_ofagent.plugins.ofagent.agent import constants as ofa_const
 from networking_ofagent.plugins.ofagent.agent import flows
+from networking_ofagent.plugins.ofagent.agent import monitor
 from networking_ofagent.plugins.ofagent.agent import ports
 from networking_ofagent.plugins.ofagent.agent import tables
 
@@ -134,6 +135,7 @@ class OFANeutronAgentRyuApp(app_manager.RyuApp):
     def __init__(self, *args, **kwargs):
         super(OFANeutronAgentRyuApp, self).__init__(*args, **kwargs)
         self.arplib = arp_lib.ArpLib(self)
+        self.monitor = monitor.PortMonitor()
 
     def start(self):
         super(OFANeutronAgentRyuApp, self).start()
@@ -166,6 +168,13 @@ class OFANeutronAgentRyuApp(app_manager.RyuApp):
     def del_arp_table_entry(self, network, ip):
         self.arplib.del_arp_table_entry(network, ip)
 
+    @handler.set_ev_cls(ofp_event.EventOFPPortStatus, handler.MAIN_DISPATCHER)
+    def _port_status_handler(self, ev):
+        self.monitor.port_status_handler(ev)
+
+    def get_port_status_list(self):
+        return self.monitor.get_port_status_list()
+
 
 class OFANeutronAgent(sg_rpc.SecurityGroupAgentRpcCallbackMixin,
                       l2population_rpc.L2populationRpcCallBackTunnelMixin):
@@ -290,6 +299,20 @@ class OFANeutronAgent(sg_rpc.SecurityGroupAgentRpcCallbackMixin,
                 self._report_state)
             heartbeat.start(interval=report_interval)
 
+    def _send_set_async(self, br):
+        """Set asynchronous configuration message for the given bridge."""
+        datapath = br.datapath
+        ofp = datapath.ofproto
+        ofpp = datapath.ofproto_parser
+        packet_in_mask = 1 << ofp.OFPR_ACTION | 1 << ofp.OFPR_INVALID_TTL
+        port_status_mask = 1 << ofp.OFPPR_DELETE
+        flow_removed_mask = 0
+        msg = ofpp.OFPSetAsync(datapath,
+                               [packet_in_mask, 0],
+                               [port_status_mask, 0],
+                               [flow_removed_mask, 0])
+        ryu_api.send_msg(app=self.ryuapp, msg=msg)
+
     def _get_ports(self, br):
         """Generate ports.Port instances for the given bridge."""
         datapath = br.datapath
@@ -312,7 +335,7 @@ class OFANeutronAgent(sg_rpc.SecurityGroupAgentRpcCallbackMixin,
             if vif_id in vlan_mapping.vif_ports:
                 return network_id
 
-    @log.log
+    @log_helpers.log_method_call
     def port_update(self, context, **kwargs):
         port = kwargs.get('port')
         # Put the port identifier in the updated_ports set.
@@ -324,7 +347,7 @@ class OFANeutronAgent(sg_rpc.SecurityGroupAgentRpcCallbackMixin,
     def _tunnel_port_lookup(self, network_type, _remote_ip):
         return self.tun_ofports.get(network_type)
 
-    @log.log
+    @log_helpers.log_method_call
     def fdb_add(self, context, fdb_entries):
         for lvm, agent_ports in self.get_agent_ports(fdb_entries,
                                                      self.local_vlan_map):
@@ -338,7 +361,7 @@ class OFANeutronAgent(sg_rpc.SecurityGroupAgentRpcCallbackMixin,
             else:
                 self._fdb_add_arp(lvm, agent_ports)
 
-    @log.log
+    @log_helpers.log_method_call
     def fdb_remove(self, context, fdb_entries):
         for lvm, agent_ports in self.get_agent_ports(fdb_entries,
                                                      self.local_vlan_map):
@@ -352,7 +375,7 @@ class OFANeutronAgent(sg_rpc.SecurityGroupAgentRpcCallbackMixin,
             else:
                 self._fdb_remove_arp(lvm, agent_ports)
 
-    @log.log
+    @log_helpers.log_method_call
     def _fdb_add_arp(self, lvm, agent_ports):
         for _remote_ip, port_infos in agent_ports.items():
             for port_info in port_infos:
@@ -362,7 +385,7 @@ class OFANeutronAgent(sg_rpc.SecurityGroupAgentRpcCallbackMixin,
                                                 port_info.ip_address,
                                                 port_info.mac_address)
 
-    @log.log
+    @log_helpers.log_method_call
     def _fdb_remove_arp(self, lvm, agent_ports):
         for _remote_ip, port_infos in agent_ports.items():
             for port_info in port_infos:
@@ -416,7 +439,7 @@ class OFANeutronAgent(sg_rpc.SecurityGroupAgentRpcCallbackMixin,
         elif action == 'remove':
             self.ryuapp.del_arp_table_entry(local_vid, ip_address)
 
-    @log.log
+    @log_helpers.log_method_call
     def _fdb_chg_ip(self, context, fdb_entries):
         self.fdb_chg_ip_tun(context, self.int_br, fdb_entries, self.local_ip,
                             self.local_vlan_map)
@@ -587,6 +610,7 @@ class OFANeutronAgent(sg_rpc.SecurityGroupAgentRpcCallbackMixin,
         br = self.int_br
         br.setup_ofp()
         br.setup_default_table()
+        self._send_set_async(br)
 
     def setup_physical_interfaces(self, interface_mappings):
         """Setup the physical network interfaces.
@@ -624,6 +648,31 @@ class OFANeutronAgent(sg_rpc.SecurityGroupAgentRpcCallbackMixin,
         port_info['removed'] = registered_ports - cur_ports
         return port_info
 
+    def _check_port_status_list(self, port_status_list, port_info):
+        """Checks if port_status_list has port-no changed ports.
+
+        Checks if there are ports which port-no change occurred in
+        list of port status messages and returns them. And adds target
+        ports to an updated set in port_info.
+
+        :param port_status_list: list of port delete status messages
+        :param port_info: dict has set of current and updated port names
+        """
+        # Currently port_status_list has only delete status messages.
+        check_ports = dict((ps.name, ps) for ps in port_status_list
+            if ps.name in port_info['current'])
+        check_ports_set = set(check_ports.keys())
+        LOG.debug("Agent daemon_loop check_ports: %(check_ports)s "
+                  "check_ports_set: %(check_ports_set)s",
+                  {'check_ports': check_ports,
+                   'check_ports_set': check_ports_set})
+        if check_ports_set:
+            if 'updated' in port_info:
+                port_info['updated'] |= check_ports_set
+            else:
+                port_info['updated'] = check_ports_set
+        return check_ports
+
     def treat_vif_port(self, vif_port, port_id, network_id, network_type,
                        physical_network, segmentation_id, admin_state_up):
         if vif_port:
@@ -681,7 +730,19 @@ class OFANeutronAgent(sg_rpc.SecurityGroupAgentRpcCallbackMixin,
         # We only have one port per network_type.
         pass
 
-    def treat_devices_added_or_updated(self, devices):
+    @log_helpers.log_method_call
+    def _repair_ofport_change(self, port, net_uuid):
+        if net_uuid not in self.local_vlan_map:
+            LOG.info(_LI("_repair_ofport_change() net_uuid %s not in "
+                         "local_vlan_map"), net_uuid)
+            return
+        lvm = self.local_vlan_map[net_uuid]
+        self.int_br.check_in_port_delete_port(port.ofport)
+        if port.vif_mac is None:
+            return
+        self.int_br.local_out_delete_port(lvm.vlan, port.vif_mac)
+
+    def treat_devices_added_or_updated(self, devices, check_ports):
         resync = False
         all_ports = dict((p.normalized_port_name(), p) for p in
                          self._get_ports(self.int_br) if p.is_neutron_port())
@@ -704,6 +765,19 @@ class OFANeutronAgent(sg_rpc.SecurityGroupAgentRpcCallbackMixin,
                           {'device': device, 'e': e})
                 resync = True
                 continue
+            if device in check_ports:
+                ps = check_ports[device]
+                if port.ofport != ps.port.ofport:
+                    ps.port.vif_mac = details.get('mac_address')
+                    LOG.debug("Repair ofport changed old port "
+                              "ofport: %(ofport)s vif_mac: %(mac)s",
+                              {'ofport': ps.port.ofport,
+                               'mac': ps.port.vif_mac})
+                    LOG.debug("Repair ofport changed new port "
+                              "ofport: %(ofport)s vif_mac: %(mac)s",
+                              {'ofport': port.ofport,
+                               'mac': port.vif_mac})
+                    self._repair_ofport_change(ps.port, details['network_id'])
             if 'port_id' in details:
                 LOG.info(_LI("Port %(device)s updated. Details: %(details)s"),
                          {'device': device, 'details': details})
@@ -749,7 +823,7 @@ class OFANeutronAgent(sg_rpc.SecurityGroupAgentRpcCallbackMixin,
             self.port_unbound(device)
         return resync
 
-    def process_network_ports(self, port_info):
+    def process_network_ports(self, port_info, check_ports):
         resync_add = False
         resync_removed = False
         # If there is an exception while processing security groups ports
@@ -766,7 +840,7 @@ class OFANeutronAgent(sg_rpc.SecurityGroupAgentRpcCallbackMixin,
         if devices_added_updated:
             start = time.time()
             resync_add = self.treat_devices_added_or_updated(
-                devices_added_updated)
+                devices_added_updated, check_ports)
             LOG.debug("process_network_ports - iteration:%(iter_num)d - "
                       "treat_devices_added_or_updated completed "
                       "in %(elapsed).3f",
@@ -838,6 +912,10 @@ class OFANeutronAgent(sg_rpc.SecurityGroupAgentRpcCallbackMixin,
                 self.updated_ports = set()
                 port_info = self.scan_ports(ports, updated_ports_copy)
                 ports = port_info['current']
+                # In the following calling port_info may be updated
+                port_status_list = self.ryuapp.get_port_status_list()
+                check_ports = self._check_port_status_list(port_status_list,
+                                                           port_info)
                 LOG.debug("Agent daemon_loop - iteration:%(iter_num)d - "
                           "port information retrieved. "
                           "Elapsed:%(elapsed).3f",
@@ -850,7 +928,7 @@ class OFANeutronAgent(sg_rpc.SecurityGroupAgentRpcCallbackMixin,
                     LOG.debug("Starting to process devices in:%s",
                               port_info)
                     # If treat devices fails - must resync with plugin
-                    sync = self.process_network_ports(port_info)
+                    sync = self.process_network_ports(port_info, check_ports)
                     LOG.debug("Agent daemon_loop - "
                               "iteration:%(iter_num)d - "
                               "ports processed. Elapsed:%(elapsed).3f",
